/*
 * check_password.c for OpenLDAP
 *
 * 2007-06-06 Michael Steinmann <msl@calivia.com>
 * Version 1.0.1
 *
 * 2008-01-30 Pierre-Yves Bonnetain <py.bonnetain@ba-cst.com>
 * Version 1.0.2
 * External configuration of quality points.
 * Use of syslog for debug traces.
 * Several bug fixes.
 */

#include <string.h>
#include <ctype.h>
#include <portable.h>
#include <slap.h>

#ifdef HAVE_CRACKLIB
#include "crack.h"
#endif

#if defined(DEBUG)
#include <syslog.h>
#endif

#ifndef CRACKLIB_DICTPATH
#define CRACKLIB_DICTPATH "/usr/share/cracklib/pw_dict"
#endif

#ifndef CONFIG_FILE
#define CONFIG_FILE "/etc/openldap/check_password.conf"
#endif

#define DEFAULT_QUALITY 3
#define MEMORY_MARGIN   50
#define MEM_INIT_SZ     64
#define FILENAME_MAXLEN 512

#define PASSWORD_TOO_SHORT_SZ \
    "Password for dn=\"%s\" is too short (%d/6)"
#define PASSWORD_QUALITY_SZ \
    "Password for dn=\"%s\" does not pass required number of strength checks (%d of %d)"
#define BAD_PASSWORD_SZ \
    "Bad password for dn=\"%s\" because %s"

typedef void (*validator) (char*, int*);
static void read_config_file (int *);
static validator valid_word (char *);
static void set_quality (char *, int *);

int check_password (char *pPasswd, char **ppErrStr, Entry *pEntry);

static void set_quality (char *value, int *quality)
{
#if defined(DEBUG)
  syslog(LOG_NOTICE, "Setting quality to [%s]", value);
#endif

  /* No need to require more quality than we can check for. */
  if (!isdigit(*value) || (int) (value[0] - '0') > 4) return;
  
  *quality = (int) (value[0] - '0');

#if defined(DEBUG)
  syslog(LOG_NOTICE, "Quality set to %d", *quality);
#endif
}

static validator valid_word (char *word)
{
  struct {
    char * parameter;
    validator dealer;
  } list[] = { { "minPoints", set_quality },
               { NULL, NULL } };
  int index = 0;

#if defined(DEBUG)
  syslog(LOG_NOTICE, "Validating parameter [%s]", word);
#endif

  while (list[index].parameter != NULL) {
    if (strlen(word) == strlen(list[index].parameter) &&
	strcmp(list[index].parameter, word) == 0) {
#if defined(DEBUG)
      syslog(LOG_NOTICE, "Parameter accepted.");
#endif
      return list[index].dealer;
    }
    index++;
  }

#if defined(DEBUG)
  syslog(LOG_NOTICE, "Parameter rejected.");
#endif

  return NULL;
}

static void read_config_file (int *quality)
{
  FILE * config;
  char * line;

  /* Set a sensible default to keep V1.0.1 behaviour. */
  *quality = DEFAULT_QUALITY;

  if ((line = ber_memcalloc(260, sizeof(char))) == NULL) {
    return;
  }

  if ( (config = fopen(CONFIG_FILE, "r")) == NULL) {
#if defined(DEBUG)
    syslog(LOG_ERR, "Opening file %s failed", CONFIG_FILE);
#endif

    ber_memfree(line);
    return;
  }

  while (fgets(line, 256, config) != NULL) {
    char *start = line;
    char *word, *value;
    validator dealer;

#if defined(DEBUG)
    /* Debug traces to syslog. */
    syslog(LOG_NOTICE, "Got line |%s|", line);
#endif

    while (isspace(*start) && isascii(*start)) start++;

    if (! isascii(*start))
      continue;

    if ((word = strtok(start, " \t")) && (dealer = valid_word(word))) {
      if ((value = strtok(NULL, " \t")) == NULL)
	continue;

#if defined(DEBUG)
      syslog(LOG_NOTICE, "Word = %s, value = %s", word, value);
#endif

      (*dealer)(value, quality);
    }
  }

  fclose(config);
  ber_memfree(line);
  return;
}

static int realloc_error_message (char ** target, int curlen, int nextlen)
{
  if (curlen < nextlen + MEMORY_MARGIN) {
#if defined(DEBUG)
    syslog(LOG_WARNING, "Reallocating szErrStr from %d to %d",
	   curlen, nextlen + MEMORY_MARGIN);
#endif
    ber_memfree(*target);
    curlen = nextlen + MEMORY_MARGIN;
    *target = (char *) ber_memalloc(curlen);
  }

  return curlen;
}

int
check_password (char *pPasswd, char **ppErrStr, Entry *pEntry)
{

    char *szErrStr = (char *) ber_memalloc(MEM_INIT_SZ);
    int  mem_len = MEM_INIT_SZ;

    int nLen;
    int nLower = 0;
    int nUpper = 0;
    int nDigit = 0;
    int nPunct = 0;
    int nQuality = 0;
    int i;
    int minQuality = 0;

    /** bail out early as cracklib will reject passwords shorter
      * than 6 characters
      */

    nLen = strlen (pPasswd);
    if ( nLen < 6) {
      mem_len = realloc_error_message(&szErrStr, mem_len,
				      strlen(PASSWORD_TOO_SHORT_SZ) +
				      strlen(pEntry->e_name.bv_val) + 1);
      sprintf (szErrStr, PASSWORD_TOO_SHORT_SZ, pEntry->e_name.bv_val, nLen);
      goto fail;
    }

    read_config_file(&minQuality);

    /** The password must have at least minQuality strength points with one
      * point for the first occurrance of a lower, upper, digit and
      * punctuation character
      */

    for ( i = 0; i < nLen; i++ ) {

        if ( nQuality >= minQuality ) break;

        if ( islower (pPasswd[i]) ) {
            if ( !nLower ) {
                nLower = 1; nQuality++;
            }
            continue;
        }

        if ( isupper (pPasswd[i]) ) {
            if ( !nUpper ) {
                nUpper = 1; nQuality++;
            }
            continue;
        }

        if ( isdigit (pPasswd[i]) ) {
            if ( !nDigit ) {
                nDigit = 1; nQuality++;
            }
            continue;
        }

        if ( ispunct (pPasswd[i]) ) {
            if ( !nPunct ) {
                nPunct = 1; nQuality++;
            }
            continue;
        }
    }

    if ( nQuality < minQuality ) {
      mem_len = realloc_error_message(&szErrStr, mem_len,
				      strlen(PASSWORD_QUALITY_SZ) +
				      strlen(pEntry->e_name.bv_val) + 2);
      sprintf (szErrStr, PASSWORD_QUALITY_SZ, pEntry->e_name.bv_val,
	       nQuality, minQuality);
      goto fail;
    }

#ifdef HAVE_CRACKLIB

    /** Check password with cracklib */

    int   j = 0;
    FILE* fp;
    char  filename[FILENAME_MAXLEN];
    char  const* ext[] = { "hwm", "pwd", "pwi" };
    int   nErr = 0;

    /**
      * Silently fail when cracklib wordlist is not found
      */

    for ( j = 0; j < 3; j++ ) {

        snprintf (filename, FILENAME_MAXLEN - 1, "%s.%s", \
		CRACKLIB_DICTPATH, ext[j]);

        if (( fp = fopen ( filename, "r")) == NULL ) {

            nErr = 1;
            break;

        } else {

            fclose (fp);

        }
    }

    char *r;
    if ( nErr  == 0) {

        r = (char *) FascistCheck (pPasswd, CRACKLIB_DICTPATH);
        if ( r != NULL ) {
	  mem_len = realloc_error_message(&szErrStr, mem_len,
					  strlen(BAD_PASSWORD_SZ) +
					  strlen(pEntry->e_name.bv_val) +
					  strlen(r));
	  sprintf (szErrStr, BAD_PASSWORD_SZ, pEntry->e_name.bv_val, r);
	  goto fail;
        }
    }

#endif

    *ppErrStr = strdup ("");
    ber_memfree(szErrStr);
    return (LDAP_SUCCESS);

fail:
    *ppErrStr = strdup (szErrStr);
    ber_memfree(szErrStr);
    return (EXIT_FAILURE);

}

